// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    intern::{Intern, Interned},
    ty::{CanonicalType, StaticType, Type},
};
use hashbrown::HashSet;
use std::ops::Index;

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct EnumVariant {
    pub name: Interned<str>,
    pub ty: Option<CanonicalType>,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Enum {
    variants: Interned<[EnumVariant]>,
}

impl Enum {
    #[track_caller]
    pub fn new(variants: &[EnumVariant]) -> Self {
        let variants = variants.intern();
        let mut names = HashSet::new();
        for variant in variants {
            assert!(
                names.insert(variant.name),
                "duplicate variant name: {:?}",
                variant.name
            );
        }
        Self { variants }
    }
}

impl Type for Enum {
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Enum(*self)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct HdlOption<T: Type> {
    some_ty: T,
}

impl<T: Type> HdlOption<T> {
    pub const fn new(some_ty: T) -> Self {
        Self { some_ty }
    }
    pub const fn some_ty(&self) -> &T {
        &self.some_ty
    }
}

#[allow(non_upper_case_globals)]
pub const HdlOption: HdlOptionWithoutGenerics = HdlOptionWithoutGenerics;

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
pub struct HdlOptionWithoutGenerics;

impl<T: Type> Index<T> for HdlOptionWithoutGenerics {
    type Output = HdlOption<T>;

    fn index(&self, some_ty: T) -> &Self::Output {
        Interned::<_>::into_inner(Intern::intern_sized(HdlOption::new(some_ty)))
    }
}

impl<T: Type> Type for HdlOption<T> {
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Enum(Enum::new(&[
            EnumVariant {
                name: "None".intern(),
                ty: None,
            },
            EnumVariant {
                name: "Some".intern(),
                ty: Some(self.some_ty().canonical()),
            },
        ]))
    }
}

impl<T: StaticType> StaticType for HdlOption<T> {
    fn static_type() -> Self {
        HdlOption::new(T::static_type())
    }
}
