// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::intern::{Intern, Interned};
use crate::ty::{CanonicalType, StaticType, Type};
use std::ops::Index;

pub const UNKNOWN_SIZE: usize = !0;

mod sealed {
    pub trait KnownSizeSealed {}
}

pub trait KnownSize<const WIDTH: usize>:
    Copy + Send + Sync + 'static + sealed::KnownSizeSealed
{
}

impl sealed::KnownSizeSealed for () {}

macro_rules! known_widths {
    ([] $($bits:literal)+) => {
        impl KnownSize<{
            let v = 0;
            $(let v = v * 2 + $bits;)*
            v
        }> for () {}
    };
    ([2 $($rest:tt)*] $($bits:literal)+) => {
        known_widths!([$($rest)*] $($bits)* 0);
        known_widths!([$($rest)*] $($bits)* 1);
    };
    ([2 $($rest:tt)*]) => {
        known_widths!([$($rest)*] 0);
        known_widths!([$($rest)*] 1);
        impl KnownSize<{2 $(* $rest)*}> for () {}
    };
}

known_widths!([2 2 2 2 2 2 2 2 2]);

macro_rules! impl_int {
    ($name:ident, $generic_name:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name<const WIDTH: usize = UNKNOWN_SIZE> {
            width: usize,
        }

        #[allow(non_upper_case_globals)]
        pub const $name: $generic_name = $generic_name;

        impl<const WIDTH: usize> $name<WIDTH> {
            pub const fn width(self) -> usize {
                if WIDTH != UNKNOWN_SIZE {
                    debug_assert!(self.width == WIDTH);
                    WIDTH
                } else {
                    self.width
                }
            }
        }

        impl $name {
            pub const fn new_dyn(width: usize) -> Self {
                assert!(width != UNKNOWN_SIZE);
                Self { width }
            }
        }

        impl<const WIDTH: usize> $name<WIDTH>
        where
            (): KnownSize<WIDTH>,
        {
            pub const fn new_static() -> Self {
                Self { width: WIDTH }
            }
        }

        impl<const WIDTH: usize> Type for $name<WIDTH> {
            fn canonical(&self) -> CanonicalType {
                CanonicalType::$name($name::new_dyn(self.width()))
            }
        }

        impl<const WIDTH: usize> StaticType for $name<WIDTH>
        where
            (): KnownSize<WIDTH>,
        {
            fn static_type() -> Self {
                Self::new_static()
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
        pub struct $generic_name;

        impl Index<usize> for $generic_name {
            type Output = $name;

            fn index(&self, width: usize) -> &Self::Output {
                Interned::<_>::into_inner(Intern::intern_sized($name::new_dyn(width)))
            }
        }
    };
}

impl_int!(UInt, UIntWithoutGenerics);
impl_int!(SInt, SIntWithoutGenerics);
